///////////////////////////////////////////////////////////////////////////////
//RevHead_Start                                                              //
//                                                                           //
// Project Details                                                           //
// ---------------                                                           //
//Project Name:    xtermsu                                                   //
//Project Version: 0.6.0                                                     //
//Copyright Owner: David M. Balean <linprogs@exemail.com.au>                 //
//Copyright Notice:                                                          //
// This program is free software; you can redistribute it and/or modify it   //
// under the terms of the GNU General Public License as published by         //
// the Free Software Foundation, either version 3 of the License or,         //
// (at your option) any later version.                                       //
// This program is distributed in the hope that it will be useful,           //
// but WITHOUT ANY WARRANTY; without even the implied warranty of            //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                      //
// See the GNU General Public License for more details.                      //
// You should have received a copy of the GNU General Public License         //
// along with this program. If not, see <http://www.gnu.org/licenses/>.      //
//                                                                           //
//Project Date:    Tuesday 12th May 2009                                     //
//                                                                           //
//Start DateStamp: 29-Apr-2009 10.42.00 -36000 0                             //
//                                                                           //
//Contact:                                                                   //
// email: linprogs@balean.net                                                //
// snail: D.M. Balean, 44 Wyong Road, Killarney Vale, N.S.W. 2261, AUSTRALIA //
//                                                                           //
//Project Summary:                                                           //
// A simple xterm frontend for the su command.                               //
// The user is allowed three attempts.                                       //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//This File:       xtermsu.cpp                                               //
//Author:          David M. Balean <linprogs@balean.net>                     //
//DateStamp:       12-May-2009 20.49.49 -36000 0                             //
//                                                                           //
//RevHead_End                                                                //
///////////////////////////////////////////////////////////////////////////////
#include <iostream>
#include <sstream>
#include <fstream>

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <pwd.h>
#include <string.h>
#include <errno.h>
#include <sys/wait.h>
#include <unistd.h>
#include <wait.h>
#include <linux/limits.h>  // for PIPE_BUF

#include "xtermsu.h"


int main(int argc, char *argv[])
{
if ( (strcmp( (getpwuid( getuid() ))->pw_name , "root" )==0)    )  
  {
  string command;
  int rc;
  for (int n=1; n<argc; ++n)
    {
    command += argv[n];
    if (n<(argc-1))
      {
      command += " ";
      }
    }
  rc = system(command.c_str());
  return rc;
  }
else
  {
  bool failed = 1;
  int trial = 0;

  for (int trial = 3; trial && failed; --trial)
    {
    Xtermsu xt;

    xt.su(argc, argv, trial);

    if (!xt.failed)
      {
      failed = 0;
      break;
      }
    }

//  if (failed) cout << "main(): FAILED" << endl;
//  else cout << "main(): SUCCESS" << endl;

  return (failed) ? EXIT_FAILURE : EXIT_SUCCESS;

  }
}


Xtermsu::Xtermsu(void)
{
failed = 1;
}


// Pipe output of a command to a buffer
string Xtermsu::get_file(char *command)
{
string buffer;
FILE *fp;
char pipebuffer[PIPE_BUF];

if ( (fp = popen(command, "r")) )
  {
  while ( fgets(pipebuffer, PIPE_BUF, fp) )
    {
    buffer += pipebuffer;
    }

  if (ferror(fp))
    {
    // ERROR CONDITION
    buffer.clear();
    }
  }
else
  {
	// ERROR: can't open pipe
  // Empty string is returned
  }
pclose(fp);
return(buffer);  // returns a zero-terminated string
}

// Open a new process, command
int Xtermsu::new_process(const char *command)  //, pid_t p)
{

extern char **environ;


pid_t pid;

// NOTE:  pipes -  write into fp[1] and read from fp[0]

if (! (command && *command) ) return(-1);

//  if (pipe(fp)) return(-2);   // NOT USING PIPES

if ((pid = fork()) == -1) return(-3);

if (pid == 0)
	{

	//==============
	// CHILD PROCESS
	//==============

  char *argv[4];
  argv[0] = (char *)"sh";
  argv[1] = (char *)"-c";
  argv[2] = (char *)command;
  argv[3] = 0;

//  cout << "CHILD PROCESS - command= " << command << endl;

  execve("/bin/sh", argv, environ);

  exit(127);                          // Error if it returns...
  }

//===============
// PARENT PROCESS
//===============

//cout << "PARENT PROCESS - command= " << command << endl;

sleep(1);    // give it a moment to start
  
// list of possible processes
char *c, *ch;
char t;
int p;

// Use   ps -eo %p %a    to pbtain pid and full command only
//       ps -eo %p %c    gives command name without args


char *plist = (char *)( get_file( (char *)"ps -eo \"%p %a\" --no-headers" ) ).c_str();

//cout << plist << endl;

if ( (c = strstr(plist, match.c_str())) )
  {
//  cout << "GOT match 1 with " << match << endl;

  while ( (*c) && ((*c)!='\n') ) c++;
  if ( (*c)='\n' ) c++;

  if ( (c = strstr(c, ps_su_command.c_str())) )
    {
//    cout << "GOT match 1 with " << ps_su_command << endl;
//    cout << "WAITING for \"" << command_x << "\n in the ps list" << endl;

    while (1)
      {
      plist = (char *)( get_file( (char *)"ps -eo \"%p %a\" --no-headers" ) ).c_str();
      c = plist;
 
      if ( (c = strstr(plist, match.c_str())) )
        {
        pid_t pid;

//cout << "GOT match 2 with " << match << endl;

        // Get the start of this line
        while (c>plist)
          {
          if ( (*(c-1))== '\n' ) break;
          c--;
          }
        while (isblank(*c)) c++;

        pid = atoi(c);


        while ( (*c) && ((*c)!='\n') ) c++;
        if ( (*c)='\n' ) c++;

        if ( (c = strstr(c, ps_su_command.c_str())) )
          {
// WE need the number of this process...

//        cout << "GOT match 2 with " << ps_su_command << endl;
          while ( (*c) && ((*c)!='\n') ) c++;
          if ( (*c)='\n' ) c++;
          
          if ( (c = strstr(c, command_x.c_str())) )
            {

              
//cout << plist << endl;
//cout << "xterm's pid is " << pid << endl;

            kill(pid, SIGTERM);
//            cout << "FOUND  match with command_x - " << command_x << endl;
            failed = 0;
            break;
            }
          }
        else
          {
//        cout << "NO match 2 with " << ps_su_command << endl;
          break;
          } 

        }
        
      else
        {
//      cout << "NO match 2 with " << match << endl;
        break;
       }

//    cout << "RETRY" << endl;
      usleep(200);
      } // END OF while
    }
  }
 
//if (failed) cout << "NO MATCH FOUND" << endl;



for(;;)
	{
  int status = 0;

  if (waitpid(pid, &status, 0) == -1)
		{
	  if (! (errno == EINTR) ) return(-1);
		}
	else if (WIFEXITED(status))
		{
	  status = (WEXITSTATUS(status));

//#ifdef debug      
//    cerr << "Leaving NewProcess("<< command << "," << p
//         << "), return code " << status << endl;     
//#endif

	  return(status);
		}
	else
		{
		
//#ifdef debug
//   cerr << "Leaving NewProcess("<< command << "," << p
//         << "), UNKNOWN ERROR: status " << status << endl;
//#endif

	  return(-4);
		}
	}
}


string Xtermsu::itos(int n)
{
ostringstream ost;
ost << n;
return ost.str();
}

// Probably best to do this with a list
void Xtermsu::su(int argc, char *argv[], int trial)
{
pid_t start_pid = getpid();

// Get a bit of info
//  command = "xterm -geometry 80x10+200+300 -wf -fn -*-lucidatypewriter-bold-r-normal-*-14-*-*-*-*-*-*-* -T \"ENTER ROOT'S PASSWORD\n\n\n";
string command = "xterm -geometry 60x10+200+300 -wf -fn -*-*-bold-r-*-*-*-180-*-*-*-*-*-* -T \"ENTER ROOT'S PASSWORD\n\n\n";
match          = "xterm -geometry 60x10+200+300 -wf -fn -*-*-bold-r-*-*-*-180-*-*-*-*-*-* -T ENTER ROOT'S PASSWORD???";
  
command += itos(start_pid);
match   += itos(start_pid);
command += "\" +sb -bg white -fg black -e \"su root -c '";

command_x.clear();
ps_su_command = "su root -c ";
su_command = "su root -c '";
for (int n=1; n<argc; n++) // forget the xtermsu (program name)
  {
  command += argv[n];
  ps_su_command += argv[n];
  su_command += argv[n];
  command_x += argv[n];
  if (n<(argc-1))
    {
    command += " ";
    ps_su_command += " ";
    su_command += " ";
    command_x += " ";
    }
  }
su_command += "'";
command += "'\" &";


//setenv("START_PID", (itos(start_pid)).c_str(), 1);


//cout << "su_command=    " << su_command << endl;
//cout << "ps_su_command= " << ps_su_command << endl;
//cout << "command_x =    " << command_x << endl;
//cout << "command   =    " << command << endl;
//cout << "match     =    " << match << endl;

//int rc =
new_process(command.c_str()); //, start_pid);

//cout << "rc= " << rc << endl;
}

